<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\AMLI\BudgetBundle\Controller;

use Chill\AMLI\BudgetBundle\Entity\AbstractElement;
use Chill\AMLI\BudgetBundle\Security\Authorization\BudgetElementVoter;
use Chill\PersonBundle\Entity\Person;
use Doctrine\ORM\EntityManagerInterface;
use Psr\Log\LoggerInterface;
use Sensio\Bundle\FrameworkExtraBundle\Configuration\Route;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\Form\Extension\Core\Type\SubmitType;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Translation\TranslatorInterface;

use function get_class;

abstract class AbstractElementController extends Controller
{
    /**
     * @var LoggerInterface
     */
    protected $chillMainLogger;

    /**
     * @var EntityManagerInterface
     */
    protected $em;

    /**
     * @var TranslatorInterface
     */
    protected $translator;

    public function __construct(
        EntityManagerInterface $em,
        TranslatorInterface $translator,
        LoggerInterface $chillMainLogger
    ) {
        $this->em = $em;
        $this->translator = $translator;
        $this->chillMainLogger = $chillMainLogger;
    }

    /**
     * Route(
     *  "{_locale}/family-members/family-members/{id}/delete",
     *  name="chill_family_members_family_members_delete"
     * ).
     *
     * @param mixed $template
     * @param mixed $flashMessage
     *
     * @return \Symfony\Component\BrowserKit\Response
     */
    protected function _delete(AbstractElement $element, Request $request, $template, $flashMessage)
    {
        $this->denyAccessUnlessGranted(BudgetElementVoter::DELETE, $element, 'You are not '
             . 'allowed to delete this family membership');

        $form = $this->createDeleteForm();

        if ($request->getMethod() === Request::METHOD_DELETE) {
            $form->handleRequest($request);

            if ($form->isValid()) {
                $this->chillMainLogger->notice('A budget element has been removed', [
                    'family_element' => get_class($element),
                    'by_user' => $this->getUser()->getUsername(),
                    'family_member_id' => $element->getId(),
                    'amount' => $element->getAmount(),
                    'type' => $element->getType(),
                ]);

                $em = $this->getDoctrine()->getManager();
                $em->remove($element);
                $em->flush();

                $this->addFlash('success', $this->translator
                    ->trans($flashMessage));

                return $this->redirectToRoute('chill_budget_elements_index', [
                    'id' => $element->getPerson()->getId(),
                ]);
            }
        }

        return $this->render($template, [
            'element' => $element,
            'delete_form' => $form->createView(),
        ]);
    }

    /**
     * @param string $template
     * @param string $flashOnSuccess
     *
     * @return \Symfony\Component\HttpFoundation\Response
     */
    protected function _edit(AbstractElement $element, Request $request, $template, $flashOnSuccess)
    {
        $this->denyAccessUnlessGranted(BudgetElementVoter::UPDATE, $element);

        $form = $this->createForm($this->getType(), $element);
        $form->add('submit', SubmitType::class);

        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $em = $this->getDoctrine()->getManager();
            $em->flush();

            $this->addFlash('success', $this->translator->trans($flashOnSuccess));

            return $this->redirectToRoute('chill_budget_elements_index', [
                'id' => $element->getPerson()->getId(),
            ]);
        }

        return $this->render($template, [
            'element' => $element,
            'form' => $form->createView(),
            'person' => $element->getPerson(),
        ]);
    }

    /**
     * @param mixed $template
     * @param mixed $flashMessageOnSuccess
     */
    protected function _new(Person $person, Request $request, $template, $flashMessageOnSuccess)
    {
        /** @var \Chill\AMLI\BudgetBundle\Entity\AbstractElement $element */
        $element = $this->createNewElement()
            ->setPerson($person);

        $this->denyAccessUnlessGranted(BudgetElementVoter::CREATE, $element);

        $form = $this->createForm($this->getType(), $element);
        $form->add('submit', SubmitType::class);

        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $em = $this->getDoctrine()->getManager();
            $em->persist($element);
            $em->flush();

            $this->addFlash('success', $this->translator->trans($flashMessageOnSuccess));

            return $this->redirectToRoute('chill_budget_elements_index', [
                'id' => $person->getId(),
            ]);
        }

        if ($form->isSubmitted()) {
            $this->addFlash('error', $this->translator->trans('This form contains errors'));
        }

        return $this->render($template, [
            'form' => $form->createView(),
            'person' => $person,
            'element' => $element,
        ]);
    }

    /**
     * Route(
     *  "{_locale}/family-members/family-members/{id}/view",
     *  name="chill_family_members_family_members_view"
     * ).
     *
     * @param mixed $template
     */
    protected function _view(AbstractElement $element, $template)
    {
        $this->denyAccessUnlessGranted(BudgetElementVoter::SHOW, $element);

        return $this->render($template, [
            'element' => $element,
        ]);
    }

    /**
     * @return AbstractElement the newly created element
     */
    abstract protected function createNewElement();

    abstract protected function getType();

    /**
     * Creates a form to delete a help request entity by id.
     *
     * @return \Symfony\Component\Form\Form The form
     */
    private function createDeleteForm()
    {
        return $this->createFormBuilder()
            ->setMethod(Request::METHOD_DELETE)
            ->add('submit', SubmitType::class, ['label' => 'Delete'])
            ->getForm();
    }
}
