<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\DocGeneratorBundle\GeneratorDriver;

use Chill\DocGeneratorBundle\GeneratorDriver\Exception\TemplateException;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\Mime\Part\DataPart;
use Symfony\Component\Mime\Part\Multipart\FormDataPart;
use Symfony\Contracts\HttpClient\Exception\DecodingExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\HttpExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\TransportExceptionInterface;
use Symfony\Contracts\HttpClient\HttpClientInterface;

class RelatorioDriver implements DriverInterface
{
    private LoggerInterface $logger;

    private HttpClientInterface $relatorioClient;

    private string $url;

    public function __construct(
        HttpClientInterface $relatorioClient,
        ParameterBagInterface $parameterBag,
        LoggerInterface $logger
    ) {
        $this->relatorioClient = $relatorioClient;
        $this->logger = $logger;
        $this->url = $parameterBag->get('chill_doc_generator')['driver']['relatorio']['url'];
    }

    public function generateFromResource($template, string $resourceType, array $data, ?string $templateName = null)
    {
        $formFields = [
            'variables' => json_encode($data),
            'template' => new DataPart($template, $templateName ?? uniqid('template_'), $resourceType),
        ];
        $form = new FormDataPart($formFields);

        try {
            $response = $this->relatorioClient->request('POST', $this->url, [
                'headers' => $form->getPreparedHeaders()->toArray(),
                'body' => $form->bodyToIterable(),
            ]);

            return $response->toStream();
        } catch (HttpExceptionInterface $e) {
            $content = $e->getResponse()->getContent(false);

            if (400 === $e->getResponse()->getStatusCode()) {
                $content = json_decode($content, true);
                $this->logger->error('relatorio: template error', [
                    'error' => $content['message'] ?? '_not defined',
                ]);

                throw new TemplateException([$content['message']]);
            }

            $this->logger->error('relatorio: error while generating document', [
                'msg' => $e->getMessage(),
                'response' => $e->getResponse()->getStatusCode(),
                'content' => $content,
            ]);

            throw $e;
        } catch (TransportExceptionInterface $e) {
            $this->logger->error('relatorio: transport exception', [
                'msg' => $e->getMessage(),
                'e' => $e->getTraceAsString(),
            ]);

            throw $e;
        } catch (DecodingExceptionInterface $e) {
            $this->logger->error('relatorio: could not decode response', [
                'msg' => $e->getMessage(),
                'e' => $e->getTraceAsString(),
            ]);

            throw $e;
        }
    }
}
