<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Phonenumber;

use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\ServerException;
use Psr\Cache\CacheItemPoolInterface;
use Psr\Log\LoggerInterface;

use function array_key_exists;
use function in_array;
use function json_decode;
use function preg_replace;
use function strlen;

/**
 * Helper to some task linked to phonenumber.
 *
 * Currently, only Twilio is supported (https://www.twilio.com/lookup). A method
 * allow to check if the helper is configured for validation. This should be used
 * before doing some validation.
 */
class PhonenumberHelper
{
    public const FORMAT_URI = 'https://lookups.twilio.com/v1/PhoneNumbers/%s';

    public const LOOKUP_URI = 'https://lookups.twilio.com/v1/PhoneNumbers/%s';

    protected CacheItemPoolInterface $cachePool;

    /**
     * TRUE if the client is properly configured.
     */
    protected bool $isConfigured = false;

    protected LoggerInterface $logger;

    /**
     * Twilio client.
     */
    protected Client $twilioClient;

    public function __construct(
        CacheItemPoolInterface $cachePool,
        $config,
        LoggerInterface $logger
    ) {
        $this->logger = $logger;
        $this->cachePool = $cachePool;

        if (
            array_key_exists('twilio_sid', $config)
            && !empty($config['twilio_sid'])
            && strlen($config['twilio_sid']) > 2
            && array_key_exists('twilio_secret', $config)
            && !empty($config['twilio_secret'])
            && strlen($config['twilio_secret']) > 2
        ) {
            $this->twilioClient = new Client([
                'auth' => [$config['twilio_sid'], $config['twilio_secret']],
            ]);
            $this->isConfigured = true;
        }
    }

    public function format($phonenumber)
    {
        return $this->performTwilioFormat($phonenumber);
    }

    /**
     * Get type (mobile, landline, ...) for phone number.
     */
    public function getType(string $phonenumber): string
    {
        return $this->performTwilioLookup($phonenumber) ?? 'unknown';
    }

    /**
     * Return true if the validation is configured and available.
     */
    public function isPhonenumberValidationConfigured(): bool
    {
        return $this->isConfigured;
    }

    /**
     * Return true if the phonenumber is a landline or voip phone.  Return always true
     * if the validation is not configured.
     *
     * @param string $phonenumber
     */
    public function isValidPhonenumberAny($phonenumber): bool
    {
        if (false === $this->isPhonenumberValidationConfigured()) {
            return true;
        }
        $validation = $this->performTwilioLookup($phonenumber);

        if (null === $validation) {
            return false;
        }

        return in_array($validation, ['landline', 'voip', 'mobile'], true);
    }

    /**
     * Return true if the phonenumber is a landline or voip phone.  Return always true
     * if the validation is not configured.
     *
     * @param string $phonenumber
     */
    public function isValidPhonenumberLandOrVoip($phonenumber): bool
    {
        if (false === $this->isPhonenumberValidationConfigured()) {
            return true;
        }

        $validation = $this->performTwilioLookup($phonenumber);

        if (null === $validation) {
            return true;
        }

        return in_array($validation, ['landline', 'voip'], true);
    }

    /**
     * REturn true if the phoennumber is a mobile phone. Return always true
     * if the validation is not configured.
     *
     * @param string $phonenumber
     */
    public function isValidPhonenumberMobile($phonenumber): bool
    {
        if (false === $this->isPhonenumberValidationConfigured()) {
            return true;
        }

        $validation = $this->performTwilioLookup($phonenumber);

        if (null === $validation) {
            return true;
        }

        return 'mobile' === $validation;
    }

    protected function performTwilioFormat($phonenumber)
    {
        if (false === $this->isPhonenumberValidationConfigured()) {
            return $phonenumber;
        }

        // filter only number
        $filtered = preg_replace('/[^0-9]/', '', $phonenumber);

        $item = $this->cachePool->getItem('pnum_format_nat_' . $filtered);

        if ($item->isHit()) {
            return $item->get();
        }

        try {
            $response = $this->twilioClient->get(sprintf(self::FORMAT_URI, '+' . $filtered), [
                'http_errors' => true,
            ]);
        } catch (ClientException $e) {
            $response = $e->getResponse();
            $this->logger->error('[phonenumber helper] Could not format number '
                . 'due to client error', [
                    'message' => $response->getBody()->getContents(),
                    'status_code' => $response->getStatusCode(),
                    'phonenumber' => $phonenumber,
                ]);

            return $phonenumber;
        } catch (ServerException $e) {
            $response = $e->getResponse();
            $this->logger->error('[phonenumber helper] Could not format number '
                . 'due to server error', [
                    'message' => $response->getBody()->getContents(),
                    'status_code' => $response->getStatusCode(),
                    'phonenumber' => $phonenumber,
                ]);

            return null;
        } catch (ConnectException $e) {
            $this->logger->error('[phonenumber helper] Could not format number '
                . 'due to connect error', [
                    'message' => $e->getMessage(),
                    'phonenumber' => $phonenumber,
                ]);

            return null;
        }

        $format = json_decode($response->getBody()->getContents())->national_format;

        $item
            ->set($format)
            // expires after 3d
            ->expiresAfter(3600 * 24 * 3);

        $this->cachePool->save($item);

        return $format;
    }

    protected function performTwilioLookup($phonenumber)
    {
        if (false === $this->isPhonenumberValidationConfigured()) {
            return null;
        }

        // filter only number
        $filtered = preg_replace('/[^0-9]/', '', $phonenumber);

        $item = $this->cachePool->getItem('pnum_' . $filtered);

        if ($item->isHit()) {
            //return $item->get();
        }

        try {
            $response = $this->twilioClient->get(sprintf(self::LOOKUP_URI, '+' . $filtered), [
                'http_errors' => true,
                'query' => [
                    'Type' => 'carrier',
                ],
            ]);
        } catch (ClientException $e) {
            return 'invalid';
        } catch (ServerException $e) {
            $response = $e->getResponse();
            $this->logger->error('[phonenumber helper] Could not perform validation '
                . 'due to server error', [
                    'message' => $response->getBody()->getContents(),
                    'status_code' => $response->getStatusCode(),
                    'phonenumber' => $phonenumber,
                ]);

            return null;
        } catch (ConnectException $e) {
            $this->logger->error('[phonenumber helper] Could not format number '
                . 'due to connect error', [
                    'message' => $e->getMessage(),
                    'phonenumber' => $phonenumber,
                ]);

            return null;
        }

        $validation = json_decode($response->getBody()->getContents())->carrier->type;

        $item
            ->set($validation)
            // expires after 12h
            ->expiresAfter(3600 * 12);

        $this->cachePool->save($item);

        return $validation;
    }
}
