<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Repository\Workflow;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Entity\Workflow\EntityWorkflow;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\QueryBuilder;
use Doctrine\Persistence\ObjectRepository;

class EntityWorkflowRepository implements ObjectRepository
{
    private EntityRepository $repository;

    public function __construct(EntityManagerInterface $entityManager)
    {
        $this->repository = $entityManager->getRepository(EntityWorkflow::class);
    }

    public function countByDest(User $user): int
    {
        $qb = $this->buildQueryByDest($user)->select('count(ew)');

        return (int) $qb->getQuery()->getSingleScalarResult();
    }

    public function countBySubscriber(User $user): int
    {
        $qb = $this->buildQueryBySubscriber($user)->select('count(ew)');

        return (int) $qb->getQuery()->getSingleScalarResult();
    }

    public function find($id): ?EntityWorkflow
    {
        return $this->repository->find($id);
    }

    /**
     * @return array|EntityWorkflow[]
     */
    public function findAll(): array
    {
        return $this->repository->findAll();
    }

    /**
     * @param null|mixed $limit
     * @param null|mixed $offset
     *
     * @return array|EntityWorkflow[]
     */
    public function findBy(array $criteria, ?array $orderBy = null, $limit = null, $offset = null): array
    {
        return $this->repository->findBy($criteria, $orderBy, $limit, $offset);
    }

    public function findByDest(User $user, ?array $orderBy = null, $limit = null, $offset = null): array
    {
        $qb = $this->buildQueryByDest($user)->select('ew');

        foreach ($orderBy as $key => $sort) {
            $qb->addOrderBy('ew.' . $key, $sort);
        }

        $qb->setMaxResults($limit)->setFirstResult($offset);

        return $qb->getQuery()->getResult();
    }

    public function findBySubscriber(User $user, ?array $orderBy = null, $limit = null, $offset = null): array
    {
        $qb = $this->buildQueryBySubscriber($user)->select('ew');

        foreach ($orderBy as $key => $sort) {
            $qb->addOrderBy('ew.' . $key, $sort);
        }

        $qb->setMaxResults($limit)->setFirstResult($offset);

        return $qb->getQuery()->getResult();
    }

    public function findOneBy(array $criteria): ?EntityWorkflow
    {
        return $this->repository->findOneBy($criteria);
    }

    public function getClassName(): string
    {
        return EntityWorkflow::class;
    }

    private function buildQueryByDest(User $user): QueryBuilder
    {
        $qb = $this->repository->createQueryBuilder('ew');

        $qb->join('ew.steps', 'step');

        $qb->where(
            $qb->expr()->andX(
                $qb->expr()->isMemberOf(':user', 'step.destUser'),
                $qb->expr()->isNull('step.transitionAfter'),
                $qb->expr()->eq('step.isFinal', "'FALSE'")
            )
        );

        $qb->setParameter('user', $user);

        return $qb;
    }

    private function buildQueryBySubscriber(User $user): QueryBuilder
    {
        $qb = $this->repository->createQueryBuilder('ew');

        $qb->where(
            $qb->expr()->orX(
                $qb->expr()->isMemberOf(':user', 'ew.subscriberToStep'),
                $qb->expr()->isMemberOf(':user', 'ew.subscriberToFinal'),
            )
        );

        $qb->setParameter('user', $user);

        return $qb;
    }
}
