import { createStore } from 'vuex';
import { datetimeToISO, ISOToDatetime, intervalDaysToISO, intervalISOToDays } from 'ChillMainAssets/chill/js/date.js';
import { findSocialActionsBySocialIssue } from 'ChillPersonAssets/vuejs/_api/SocialWorkSocialAction.js';
import { create } from 'ChillPersonAssets/vuejs/_api/AccompanyingCourseWork.js';
import { fetchResults, makeFetch } from 'ChillMainAssets/lib/api/apiMethods.js';
import { fetchTemplates } from 'ChillDocGeneratorAssets/api/pickTemplate.js';

const debug = process.env.NODE_ENV !== 'production';
const evalFQDN = encodeURIComponent("Chill\\PersonBundle\\Entity\\AccompanyingPeriod\\AccompanyingPeriodWorkEvaluation");

const store = createStore({
  strict: debug,
  state: {
    work: window.accompanyingCourseWork,
    startDate: window.accompanyingCourseWork.startDate !== null ?
        ISOToDatetime(window.accompanyingCourseWork.startDate.datetime) : null,
    endDate: window.accompanyingCourseWork.endDate !== null ?
        ISOToDatetime(window.accompanyingCourseWork.endDate.datetime) : null,
    note: window.accompanyingCourseWork.note,
    goalsPicked: window.accompanyingCourseWork.goals,
    goalsForAction: [],
    resultsPicked: window.accompanyingCourseWork.results,
    resultsForAction: [],
    resultsForGoal: [],
    evaluationsPicked: [],
    evaluationsForAction: [],
    templatesAvailablesForAction: [],
    templatesAvailablesForEvaluation: new Map([]),
    personsPicked: window.accompanyingCourseWork.persons,
    personsReachables: window.accompanyingCourseWork.accompanyingPeriod.participations.filter(p => p.endDate == null)
      .map(p => p.person),
    handlingThirdParty: window.accompanyingCourseWork.handlingThierParty,
    thirdParties: window.accompanyingCourseWork.thirdParties,
    isPosting: false,
    errors: [],
  },
  getters: {
    socialAction(state) {
      return state.work.socialAction;
    },
    hasResultsForAction(state) {
      return state.resultsForAction.length > 0;
    },
    resultsForGoal: (state) => (goal) => {
      let founds = state.resultsForGoal.filter(r => r.goalId === goal.id);

      return founds === undefined ? [] : founds;
    },
    resultsPickedForGoal: (state) => (goal) => {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      return found === undefined ? [] : found.results;
    },
    hasHandlingThirdParty(state) {
      return state.handlingThirdParty !== null;
    },
    hasThirdParties(state) {
      return state.thirdParties.length > 0;
    },
    getTemplatesAvailablesForEvaluation: (state) => (evaluation) => {
      if (state.templatesAvailablesForEvaluation.has(evaluation.id)) {
        return state.templatesAvailablesForEvaluation.get(evaluation.id);
      }

      return [];
    },
    buildPayload(state) {
      return {
        type: 'accompanying_period_work',
        id: state.work.id,
        startDate: state.startDate === null ? null : {
          datetime: datetimeToISO(state.startDate)
        },
        endDate: state.endDate === null ? null : {
          datetime: datetimeToISO(state.endDate)
        },
        note: state.note,
        persons: state.personsPicked.map(p => ({id: p.id, type: p.type})),
        handlingThierParty: state.handlingThirdParty === null ? null : {
          id: state.handlingThirdParty.id,
          type: state.handlingThirdParty.type
        },
        results: state.resultsPicked.map(r => ({id: r.id, type: r.type})),
        thirdParties: state.thirdParties.map(t => ({id: t.id, type: t.type})),
        goals: state.goalsPicked.map(g => {
          let o = {
            type: g.type,
            note: g.note,
            goal: {
              type: g.goal.type,
              id: g.goal.id,
            },
            results: g.results.map(r => ({id: r.id, type: r.type})),
          };
          if (g.id !== undefined) {
            o.id = g.id;
          }
          return o;
        }),
        accompanyingPeriodWorkEvaluations: state.evaluationsPicked.map(e => {
          let o = {
            type: e.type,
            key: e.key,
            evaluation: {
              id: e.evaluation.id,
              type: e.evaluation.type
            },
            startDate: e.startDate !== null ? { datetime: datetimeToISO(e.startDate) } : null,
            endDate: e.endDate !== null ? { datetime: datetimeToISO(e.endDate) } : null,
            maxDate: e.maxDate !== null ? { datetime: datetimeToISO(e.maxDate) } : null,
            warningInterval: intervalDaysToISO(e.warningInterval),
            comment: e.comment,
          };
          if (e.id !== undefined) {
            o.id = e.id;
          }

          return o;
        })
      };
    }
  },
  mutations: {
    setEvaluationsPicked(state, evaluations) {
      state.evaluationsPicked = evaluations.map((e, index) => {
        var k = Object.assign(e, {
          key: index,
          editEvaluation: false,
          startDate: e.startDate !== null ? ISOToDatetime(e.startDate.datetime) : null,
          endDate: e.endDate !== null ? ISOToDatetime(e.endDate.datetime) : null,
          maxDate: e.maxDate !== null ? ISOToDatetime(e.maxDate.datetime) : null,
          warningInterval: e.warningInterval !== null ? intervalISOToDays(e.warningInterval) : null,
        });

        return k;
      });
    },
    setStartDate(state, date) {
      state.startDate = date;
    },
    setEndDate(state, date) {
      state.endDate = date;
    },
    setResultsForAction(state, results) {
      state.resultsForAction = results;
    },
    setResultsForGoal(state, { goal, results }) {
      state.goalsForAction.push(goal);
      for (let i in results) {
        let r = results[i];
        r.goalId = goal.id;
        state.resultsForGoal.push(r);
      }
    },
    setEvaluationsForAction(state, results) {
      state.evaluationsForAction = results;
    },
    addResultPicked(state, result) {
      state.resultsPicked.push(result);
    },
    removeResultPicked(state, result) {
      state.resultsPicked = state.resultsPicked.filter(r => r.id !== result.id);
    },
    addGoal(state, goal) {
      let g = {
        type: "accompanying_period_work_goal",
        goal: goal,
        note: '',
        results: []
      }
      const tmpIndex = () => {
        let ar = state.goalsPicked.map(g => g.id),
            s = Math.min(...ar);
        return (s < 0) ? s : 0
      };
      g.id = tmpIndex() -1
      state.goalsPicked.push(g);
    },
    removeGoal(state, goal) {
      state.goalsPicked = state.goalsPicked.filter(g => g.id !== goal.id);
    },
    addResultForGoalPicked(state, { goal, result}) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      if (found === undefined) {
        return;
      }

      found.results.push(result);
    },
    removeResultForGoalPicked(state, { goal, result }) {
      let found = state.goalsPicked.find(g => g.goal.id === goal.id);

      if (found === undefined) {
        return;
      }

      found.results = found.results.filter(r => r.id !== result.id);
    },
    addEvaluation(state, evaluation) {
      let e = {
        type: "accompanying_period_work_evaluation",
        key: state.evaluationsPicked.length + 1,
        evaluation: evaluation,
        startDate: new Date(),
        endDate:  null,
        maxDate:  null,
        warningInterval: null,
        comment: "",
        editEvaluation: true,
        workflows_availables: state.work.workflows_availables_evaluation,
        documents: [],
        workflows: [],
      };
      state.evaluationsPicked.push(e);
    },
    removeEvaluation(state, evaluation) {
      state.evaluationsPicked = state.evaluationsPicked.filter(e => e.key !== evaluation.key);
    },
    setEvaluationStartDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .startDate = date;
    },
    setEvaluationEndDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .endDate = date;
    },
    setEvaluationMaxDate(state, {key, date}) {
      state.evaluationsPicked.find(e => e.key === key)
          .maxDate = date;
    },
    setEvaluationWarningInterval(state, {key, days}) {
      state.evaluationsPicked.find(e => e.key === key)
          .warningInterval = days;
    },
    setEvaluationComment(state, {key, comment}) {
      state.evaluationsPicked.find(e => e.key === key)
          .comment = comment;
    },
    toggleEvaluationEdit(state, {key}) {
      let evaluation = state.evaluationsPicked.find(e => e.key === key);
      evaluation.editEvaluation = !evaluation.editEvaluation;
    },
    setTemplatesForEvaluation(state, {templates, evaluation}) {
      state.templatesAvailablesForEvaluation.set(evaluation.id, templates);
    },
    setTemplatesAvailablesForAction(state, templates) {
      state.templatesAvailablesForAction = templates;
    },
    setPersonsPickedIds(state, ids) {
      state.personsPicked = state.personsReachables
        .filter(p => ids.includes(p.id))
    },
    setNote(state, note) {
      state.note = note;
    },
    setHandlingThirdParty(state, thirdParty) {
      state.handlingThirdParty = thirdParty;
    },
    addThirdParties(state, thirdParties) {
      // filter to remove existing thirdparties
      let ids = state.thirdParties.map(t => t.id);
      let unexistings = thirdParties.filter(t => !ids.includes(t.id));

      for (let i in unexistings) {
        state.thirdParties.push(unexistings[i]);
      }
    },
    updateThirdParty(state, thirdParty) {
      for (let t of state.thirdParties) {
        if (t.id === thirdParty.id){
          state.thirdParties = state.thirdParties.filter(t => t.id !== thirdParty.id);
          state.thirdParties.push(thirdParty);
        }
      }
    },
    removeThirdParty(state, thirdParty) {
      state.thirdParties = state.thirdParties
        .filter(t => t.id !== thirdParty.id);
    },
    setErrors(state, errors) {
      state.errors = errors;
    },
    setIsPosting(state, st) {
      state.isPosting = st;
    },
  },
  actions: {
    updateThirdParty({ commit }, payload) {
      console.log(payload);
      commit('updateThirdParty', payload);
    },
    getReachablesGoalsForAction({ getters, commit, dispatch }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/goal/by-social-action/${socialActionId}.json`
      ;
      window
        .fetch(
          url
        ).then( response => {
          if (response.ok) {
            return response.json();
          }
          throw { m: 'Error while retriving goal for social action', s: response.status, b: response.body };
        }).then( data => {
           for (let i in data.results) {
             dispatch('getReachablesResultsForGoal', data.results[i]);
           }
        }).catch( errors => {
            commit('addErrors', errors);
        });
    },
    getReachablesResultsForGoal({ commit }, goal) {
      let
        url = `/api/1.0/person/social-work/result/by-goal/${goal.id}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for goal', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setResultsForGoal', { goal, results: data.results });
        });
    },
    getReachablesResultsForAction({ getters, commit }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/result/by-social-action/${socialActionId}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }

          throw { m: 'Error while retriving results for social action', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setResultsForAction', data.results);
        });
    },
    getReachablesEvaluationsForAction({ getters, commit }) {
      let
        socialActionId = getters.socialAction.id,
        url = `/api/1.0/person/social-work/evaluation/by-social-action/${socialActionId}.json`
      ;
      window.fetch(url)
        .then(response => {
          if (response.ok) {
            return response.json();
          }
          throw { m: 'Error while retriving evaluations for social action', s: response.status, b: response.body };
        })
        .then(data => {
          commit('setEvaluationsForAction', data.results);
        });
    },
    addEvaluation({commit, dispatch}, evaluation) {
      commit('addEvaluation', evaluation);
      dispatch('fetchTemplatesAvailablesForEvaluation', evaluation);
    },
    fetchTemplatesAvailablesForEvaluation({commit, state}, evaluation) {
      if (!state.templatesAvailablesForEvaluation.has(evaluation.id)) {
        // commit an empty array to avoid parallel fetching for same evaluation id
        commit('setTemplatesForEvaluation', {templates: [], evaluation});
        fetchResults(`/api/1.0/person/docgen/template/by-evaluation/${evaluation.id}.json`)
          .then(templates => {
            commit('setTemplatesForEvaluation', {templates, evaluation});
          });
      }
    },
    submit({ getters, state, commit }, callback) {
      let
        payload = getters.buildPayload,
        url = `/api/1.0/person/accompanying-course/work/${state.work.id}.json`,
        errors = []
      ;

      commit('setIsPosting', true);

      return makeFetch('PUT', url, payload)
        .then(data => {
          if (typeof(callback) !== 'undefined') {
            return callback(data);
          } else {
            console.info('nothing to do here, bye bye');
            window.location.assign(`/fr/person/accompanying-period/${state.work.accompanyingPeriod.id}/work`);
          }
        }).catch(error => {
          console.log('error on submit', error);
          commit('setIsPosting', false);
          commit('setErrors', error.violations);
        });
    },
    patchOnTheFly({ commit }, payload) {
      let body = { type: payload.type };
      const id = payload.data.id;
      let url = `/api/1.0/person/person/${id}.json`;
      let mutation = "updatePerson";

      if (payload.type === 'person') {
          body.firstName = payload.data.firstName;
          body.lastName = payload.data.lastName;
          if (payload.data.birthdate !== null) { body.birthdate = payload.data.birthdate; }
            body.phonenumber = payload.data.phonenumber;
            body.mobilenumber = payload.data.mobilenumber;
            body.gender = payload.data.gender;
      } else if (payload.type === 'thirdparty') {
          body.name = payload.data.text;
          body.email = payload.data.email;
          body.telephone = payload.data.phonenumber;
          body.address = { id: payload.data.address.address_id };

        url = `/api/1.0/thirdparty/thirdparty/${id}.json`;
        mutation = 'updateThirdparty'
      }
      makeFetch('PATCH', url, body)
        .then((response) => {
            commit(mutation, {target: payload.target, thirdparty: response});
        })
        .catch((error) => {
            throw error;
        })
  },
  }
});

store.commit('setEvaluationsPicked', window.accompanyingCourseWork.accompanyingPeriodWorkEvaluations);
store.dispatch('getReachablesResultsForAction');
store.dispatch('getReachablesGoalsForAction');
store.dispatch('getReachablesEvaluationsForAction');

store.state.evaluationsPicked.forEach(evaluation => {
  store.dispatch('fetchTemplatesAvailablesForEvaluation', evaluation.evaluation)
});

fetchTemplates('Chill\\PersonBundle\\Entity\\AccompanyingPeriod\\AccompanyingPeriodWork')
  .then(templates => {
    store.commit('setTemplatesAvailablesForAction', templates);
  }
)

export { store };
