<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Security\Authorization;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Security\Authorization\AbstractChillVoter;
use Chill\MainBundle\Security\Authorization\VoterHelperFactoryInterface;
use Chill\MainBundle\Security\Authorization\VoterHelperInterface;
use Chill\MainBundle\Security\ProvideRoleHierarchyInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\Person;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Security;

use function in_array;

class AccompanyingPeriodVoter extends AbstractChillVoter implements ProvideRoleHierarchyInterface
{
    public const ALL = [
        self::SEE,
        self::SEE_DETAILS,
        self::CREATE,
        self::EDIT,
        self::DELETE,
        self::FULL,
        self::TOGGLE_CONFIDENTIAL_ALL,
    ];

    public const CREATE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_CREATE';

    public const DELETE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_DELETE';

    public const EDIT = 'CHILL_PERSON_ACCOMPANYING_PERIOD_UPDATE';

    /**
     * Give all the right above.
     */
    public const FULL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_FULL';

    public const SEE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_SEE';

    /**
     * details are for seeing:.
     *
     * * SocialIssues
     */
    public const SEE_DETAILS = 'CHILL_PERSON_ACCOMPANYING_PERIOD_SEE_DETAILS';

    public const TOGGLE_CONFIDENTIAL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_TOGGLE_CONFIDENTIAL';

    /**
     * Right to toggle confidentiality.
     */
    public const TOGGLE_CONFIDENTIAL_ALL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_TOGGLE_CONFIDENTIAL_ALL';

    private Security $security;

    private VoterHelperInterface $voterHelper;

    public function __construct(
        Security $security,
        VoterHelperFactoryInterface $voterHelperFactory
    ) {
        $this->security = $security;
        $this->voterHelper = $voterHelperFactory
            ->generate(self::class)
            ->addCheckFor(null, [self::CREATE])
            ->addCheckFor(AccompanyingPeriod::class, [self::TOGGLE_CONFIDENTIAL, ...self::ALL])
            ->addCheckFor(Person::class, [self::SEE])
            ->build();
    }

    public function getRoles(): array
    {
        return self::ALL;
    }

    public function getRolesWithHierarchy(): array
    {
        return ['Accompanying period' => $this->getRoles()];
    }

    public function getRolesWithoutScope(): array
    {
        return [];
    }

    protected function supports($attribute, $subject)
    {
        return $this->voterHelper->supports($attribute, $subject);
    }

    protected function voteOnAttribute($attribute, $subject, TokenInterface $token): bool
    {
        if (!$token->getUser() instanceof User) {
            return false;
        }

        if ($subject instanceof AccompanyingPeriod) {
            if (AccompanyingPeriod::STEP_CLOSED === $subject->getStep()) {
                if (in_array($attribute, [self::EDIT, self::DELETE], true)) {
                    return false;
                }
            }

            if (AccompanyingPeriod::STEP_DRAFT === $subject->getStep()) {
                // only creator can see, edit, delete, etc.
                if (
                    $subject->getCreatedBy() === $token->getUser()
                    || null === $subject->getCreatedBy()
                ) {
                    return true;
                }

                return false;
            }

            if (self::TOGGLE_CONFIDENTIAL === $attribute) {
                if ($subject->getUser() === $token->getUser()) {
                    return true;
                }

                return $this->voterHelper->voteOnAttribute(self::TOGGLE_CONFIDENTIAL_ALL, $subject, $token);
            }

            // if confidential, only the referent can see it
            if ($subject->isConfidential()) {
                return $token->getUser() === $subject->getUser();
            }
        }

        return $this->voterHelper->voteOnAttribute($attribute, $subject, $token);
    }
}
