<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Tests\Entity;

use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\AccompanyingPeriod\Comment;
use Chill\PersonBundle\Entity\AccompanyingPeriodParticipation;
use Chill\PersonBundle\Entity\Person;
use Chill\ThirdPartyBundle\Entity\ThirdParty;
use DateTime;
use DateTimeInterface;
use function count;

/**
 * @internal
 * @coversNothing
 */
final class AccompanyingPeriodTest extends \PHPUnit\Framework\TestCase
{
    public function testClosingEqualOpening()
    {
        $datetime = new DateTime('now');

        $period = new AccompanyingPeriod($datetime);
        $period->setClosingDate($datetime);

        $this->assertTrue($period->isClosingAfterOpening());
    }

    public function testClosingIsAfterOpeningConsistency()
    {
        $datetime1 = new DateTime('now');
        $datetime2 = new DateTime('tomorrow');

        $period = new AccompanyingPeriod($datetime1);
        $period->setClosingDate($datetime2);

        $r = $period->isClosingAfterOpening();

        $this->assertTrue($r);
    }

    public function testClosingIsBeforeOpeningConsistency()
    {
        $datetime1 = new DateTime('tomorrow');
        $datetime2 = new DateTime('now');

        $period = new AccompanyingPeriod($datetime1);
        $period->setClosingDate($datetime2);

        $this->assertFalse($period->isClosingAfterOpening());
    }

    public function testIsClosed()
    {
        $period = new AccompanyingPeriod(new DateTime());
        $period->setClosingDate(new DateTime('tomorrow'));

        $this->assertFalse($period->isOpen());
    }

    public function testIsOpen()
    {
        $period = new AccompanyingPeriod(new DateTime());

        $this->assertTrue($period->isOpen());
    }

    public function testPersonPeriod()
    {
        $person = new Person();
        $person2 = new Person();
        $person3 = new Person();
        $period = new AccompanyingPeriod(new DateTime());

        $participation0 = $period->createParticipationFor($person);
        $period->createParticipationFor($person2);
        $period->createParticipationFor($person3);

        $this->assertNotNull($participation0);
        $this->assertEquals(3, $period->getParticipations()->count());
        $this->assertTrue($period->containsPerson($person));
        $this->assertFalse($period->containsPerson(new Person()));

        $participation = $period->getOpenParticipationContainsPerson($person);
        $participations = $period->getParticipationsContainsPerson($person);
        $this->assertNotNull($participation);
        $this->assertSame($person, $participation->getPerson());
        $this->assertSame($participation, $participation0);
        $this->assertEquals(1, $participations->count());

        $participationL = $period->closeParticipationFor($person);
        $this->assertSame($participationL, $participation);
        $this->assertTrue($participation->getEndDate() instanceof DateTimeInterface);

        $participation = $period->getOpenParticipationContainsPerson($person);
        $this->assertNull($participation);

        $person4 = new Person();
        $participations4 = $period->getParticipationsContainsPerson($person4);
        $this->assertEquals(0, $participations4->count());
        $participation4 = $period->getOpenParticipationContainsPerson($person4);
        $this->assertNull($participation4);

        $period->addPerson($person4);
        $this->assertInstanceOf(AccompanyingPeriodParticipation::class, $period->getOpenParticipationContainsPerson($person4));
        $this->assertEquals(1, $period->getParticipationsContainsPerson($person4)->count());
        $period->removePerson($person4);
        $this->assertNull($period->getOpenParticipationContainsPerson($person4));
        $this->assertEquals(1, $period->getParticipationsContainsPerson($person4)->count());
    }

    public function testPinnedComment()
    {
        $period = new AccompanyingPeriod(new DateTime());
        $comment = new Comment();
        $replacingComment = new Comment();

        $period->setPinnedComment(null);
        $this->assertNull($period->getPinnedComment());

        $period->setPinnedComment($comment);

        $this->assertSame($period->getPinnedComment(), $comment);
        $this->assertNull($comment->getAccompanyingPeriod());
        $this->assertEquals(0, count($period->getComments()));

        $period->setPinnedComment($replacingComment);

        $this->assertSame($period->getPinnedComment(), $replacingComment);
        $this->assertNull($replacingComment->getAccompanyingPeriod());
        $this->assertSame($period, $comment->getAccompanyingPeriod());
        $this->assertEquals(1, count($period->getComments()));
        $this->assertContains($comment, $period->getComments());

        $period->setPinnedComment(null);

        $this->assertNull($period->getPinnedComment());
        $this->assertSame($period, $comment->getAccompanyingPeriod());
        $this->assertSame($period, $replacingComment->getAccompanyingPeriod());
        $this->assertEquals(2, count($period->getComments()));
        $this->assertContains($comment, $period->getComments());
        $this->assertContains($replacingComment, $period->getComments());
    }

    public function testRequestor()
    {
        $period = new AccompanyingPeriod(new DateTime());
        $person = new Person();
        $thirdParty = new ThirdParty();

        $this->assertNull($period->getRequestorThirdParty());
        $this->assertNull($period->getRequestorPerson());
        $this->assertNull($period->getRequestor());

        $period->setRequestor($person);
        $this->assertNull($period->getRequestorThirdParty());
        $this->assertSame($person, $period->getRequestorPerson());
        $this->assertSame($person, $period->getRequestor());

        $period->setRequestor($thirdParty);
        $this->assertNull($period->getRequestorPerson());
        $this->assertSame($thirdParty, $period->getRequestorThirdParty());
        $this->assertSame($thirdParty, $period->getRequestor());

        $period->setRequestor(null);
        $this->assertNull($period->getRequestorThirdParty());
        $this->assertNull($period->getRequestorPerson());
        $this->assertNull($period->getRequestor());
    }
}
