<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\WopiBundle\Service\Wopi;

use ChampsLibres\AsyncUploaderBundle\TempUrl\TempUrlGeneratorInterface;
use ChampsLibres\WopiLib\Contract\Entity\Document;
use ChampsLibres\WopiLib\Contract\Service\DocumentLockManagerInterface;
use ChampsLibres\WopiLib\Contract\Service\DocumentManagerInterface;
use Chill\DocStoreBundle\Entity\StoredObject;
use Chill\DocStoreBundle\Repository\StoredObjectRepository;
use DateTimeInterface;
use Doctrine\ORM\EntityManagerInterface;
use Error;
use loophp\psr17\Psr17Interface;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\StreamInterface;
use Ramsey\Uuid\Uuid;
use Symfony\Bridge\PsrHttpMessage\HttpMessageFactoryInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Mime\MimeTypes;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Contracts\HttpClient\HttpClientInterface;
use Throwable;

use function strlen;

final class ChillDocumentManager implements DocumentManagerInterface
{
    private DocumentLockManagerInterface $documentLockManager;

    private EntityManagerInterface $entityManager;

    private HttpClientInterface $httpClient;

    private Psr17Interface $psr17;

    private RequestInterface $request;

    private StoredObjectRepository $storedObjectRepository;

    private TempUrlGeneratorInterface $tempUrlGenerator;

    public function __construct(
        DocumentLockManagerInterface $documentLockManager,
        EntityManagerInterface $entityManager,
        HttpClientInterface $httpClient,
        Psr17Interface $psr17,
        StoredObjectRepository $storedObjectRepository,
        TempUrlGeneratorInterface $tempUrlGenerator,
        HttpMessageFactoryInterface $httpMessageFactory,
        RequestStack $requestStack
    ) {
        $this->entityManager = $entityManager;
        $this->psr17 = $psr17;
        $this->storedObjectRepository = $storedObjectRepository;
        $this->documentLockManager = $documentLockManager;
        $this->tempUrlGenerator = $tempUrlGenerator;
        $this->httpClient = $httpClient;
        $this->request = $httpMessageFactory->createRequest($requestStack->getCurrentRequest());
    }

    public function create(array $data): Document
    {
        /** @var StoredObject $document */
        $document = (new ObjectNormalizer())->denormalize([], StoredObject::class);

        // Mime types / extension handling.
        $mimeTypes = new MimeTypes();
        $mimeTypes->getMimeTypes($data['extension']);
        $document->setType(reset($mimeTypes));

        $document->setFilename($data['name']);

        $this->entityManager->persist($document);
        $this->entityManager->flush($document);

        // TODO : Ask proper mapping.
        // Available: basename, name, extension, content, size
        $this->setContent($document, $data['content']);

        return $document;
    }

    public function deleteLock(Document $document): void
    {
        $this->documentLockManager->deleteLock($document, $this->request);
    }

    /**
     * @param string $documentFilename without extension !
     */
    public function findByDocumentFilename(string $documentFilename): ?Document
    {
        return $this->storedObjectRepository->findOneBy(
            [
                'filename' => $documentFilename,
            ]
        );
    }

    public function findByDocumentId(string $documentId): ?Document
    {
        return $this->storedObjectRepository->findOneBy(
            [
                'uuid' => Uuid::fromString($documentId),
            ]
        );
    }

    /**
     * @param StoredObject $document
     *
     * @return string The document filename with its extension.
     */
    public function getBasename(Document $document): string
    {
        $exts = (new MimeTypes())->getExtensions($document->getType());

        if ([] === $exts) {
            throw new Error('Unknown mimetype for stored document.');
        }

        return sprintf('%s.%s', $document->getFilename(), reset($exts));
    }

    /**
     * @param StoredObject $document
     */
    public function getCreationDate(Document $document): DateTimeInterface
    {
        return $document->getCreationDate();
    }

    /**
     * @param StoredObject $document
     */
    public function getDocumentId(Document $document): string
    {
        return (string) $document->getUuid();
    }

    /**
     * @param StoredObject $document
     */
    public function getLastModifiedDate(Document $document): DateTimeInterface
    {
        // TODO: Add column 'LastModifiedDate' in StoredObject entity
        return $document->getCreationDate();
    }

    public function getLock(Document $document): string
    {
        return $this->documentLockManager->getLock($document, $this->request);
    }

    public function getSha256(Document $document): string
    {
        return base64_encode(hash('sha256', $this->getContent($document)));
    }

    public function getSize(Document $document): int
    {
        return strlen($this->getContent($document));
    }

    public function getVersion(Document $document): string
    {
        // TODO ?
        return '0';
    }

    public function hasLock(Document $document): bool
    {
        return $this->documentLockManager->hasLock($document, $this->request);
    }

    public function lock(Document $document, string $lock): void
    {
        $this->documentLockManager->setLock($document, $lock, $this->request);
    }

    public function read(Document $document): StreamInterface
    {
        return $this
            ->psr17
            ->createStream($this->getContent($document));
    }

    public function remove(Document $document): void
    {
        $entityIsDeleted = false;

        try {
            $this->entityManager->remove($document);
            $entityIsDeleted = true;
        } catch (Throwable $e) {
            $entityIsDeleted = false;
        }

        if (true === $entityIsDeleted) {
            $this->deleteContent($document);
        }
    }

    public function write(Document $document, array $properties = []): void
    {
        $this->setContent($document, $properties['content']);
    }

    private function deleteContent(StoredObject $storedObject): void
    {
        /** @var StdClass $object */
        $object = $this->tempUrlGenerator->generate('DELETE', $storedObject->getFilename());

        $response = $this->httpClient->request('DELETE', $object->url);

        if (200 !== $response->getStatusCode()) {
            throw new Error('Unable to delete stored object.');
        }
    }

    private function getContent(StoredObject $storedObject): string
    {
        /** @var StdClass $object */
        $object = $this->tempUrlGenerator->generate('GET', $storedObject->getFilename());

        $response = $this->httpClient->request('GET', $object->url);

        if (200 !== $response->getStatusCode()) {
            throw new Error('Unable to retrieve stored object.');
        }

        return $response->getContent();
    }

    private function setContent(StoredObject $storedObject, string $content): void
    {
        // TODO: Add strict typing in champs-libres/async-uploader-bundle
        /** @var StdClass $object */
        $object = $this->tempUrlGenerator->generate('PUT', $storedObject->getFilename());

        $response = $this->httpClient->request('PUT', $object->url, ['body' => $content]);

        if (201 !== $response->getStatusCode()) {
            throw new Error('Unable to save stored object.');
        }
    }
}
